From f875bc54d29dc60cfdd1ea71a1432d01a1f3225d Mon Sep 17 00:00:00 2001
From: Shugo Maeda <shugo@ruby-lang.org>
Date: Mon, 18 Oct 2010 00:06:12 +0900
Subject: [PATCH 26/51] supported classbox inheritance.

---
 class.c               |    8 ++++++++
 include/ruby/intern.h |    1 +
 insns.def             |   12 +++++++-----
 lib/classbox.rb       |   34 ++++++++++++++++++++++++++++++++++
 object.c              |    1 +
 5 files changed, 51 insertions(+), 5 deletions(-)

diff --git a/class.c b/class.c
index 9a36efb..3140f67 100644
--- a/class.c
+++ b/class.c
@@ -418,6 +418,14 @@ rb_define_class_id(ID id, VALUE super)
     return klass;
 }
 
+VALUE
+rb_mod_opened(VALUE mod)
+{
+    ID opened;
+    CONST_ID(opened, "__opened__");
+    return rb_funcall(mod, opened, 0);
+}
+
 
 /*!
  * Calls Class#inherited.
diff --git a/include/ruby/intern.h b/include/ruby/intern.h
index e417feb..d2cb771 100644
--- a/include/ruby/intern.h
+++ b/include/ruby/intern.h
@@ -164,6 +164,7 @@ void rb_singleton_class_attached(VALUE,VALUE);
 VALUE rb_make_metaclass(VALUE, VALUE);
 void rb_check_inheritable(VALUE);
 VALUE rb_class_inherited(VALUE, VALUE);
+VALUE rb_mod_opened(VALUE);
 VALUE rb_define_class_id(ID, VALUE);
 VALUE rb_define_class_id_under(VALUE, ID, VALUE);
 VALUE rb_module_new(void);
diff --git a/insns.def b/insns.def
index 00a9566..c04b215 100644
--- a/insns.def
+++ b/insns.def
@@ -963,14 +963,15 @@ defineclass
 	vm_check_if_namespace(cbase);
 
 	/* find klass */
-#if 0
 	if (rb_const_defined_at(cbase, id)) {
 	    klass = rb_const_get_at(cbase, id);
-	    rb_raise(rb_eNameError, "%s is not a module", rb_id2name(id));
+	    /* already exist */
+	    if (TYPE(klass) != T_MODULE ||
+		!rb_obj_is_kind_of(klass, rb_cClassbox)) {
+		rb_raise(rb_eTypeError, "%s is not a classbox", rb_id2name(id));
+	    }
 	}
-	else
-#endif
-	{
+	else {
 	    /* new module declaration */
 	    klass = rb_define_classbox_id(id);
 	    rb_set_class_path_string(klass, cbase, rb_id2str(id));
@@ -988,6 +989,7 @@ defineclass
 		  VM_FRAME_MAGIC_CLASS, klass, 0, (VALUE) GET_BLOCK_PTR(),
 		  class_iseq->iseq_encoded, GET_SP(), 0,
 		  class_iseq->local_size);
+    rb_mod_opened(klass);
     RESTORE_REGS();
 
     INC_VM_STATE_VERSION();
diff --git a/lib/classbox.rb b/lib/classbox.rb
index 67bf1b1..d1623f9 100644
--- a/lib/classbox.rb
+++ b/lib/classbox.rb
@@ -1,5 +1,6 @@
 class Classbox
   def import_to(binding)
+    return unless @__overlayed_modules__
     @__overlayed_modules__.each do |klass, modules|
       modules.each do |mod|
         overlay_module(klass, mod, binding)
@@ -41,3 +42,36 @@ end
 def import(klassbox)
   klassbox.import_to(binding(1))
 end
+
+class Module
+  def import(klassbox)
+    @__imported_classboxes__ ||= {}
+    @__imported_classboxes__[klassbox] = true
+    klassbox.import_to(binding(1))
+  end
+
+  def __opened__(b = binding(1))
+    return unless @__imported_classboxes__
+    @__imported_classboxes__.each_key do |klassbox|
+      klassbox.import_to(b)
+    end
+  end
+end
+
+class Class
+  def __opened__
+    unless @__imported_classboxes__
+      @__imported_classboxes__ =
+        self.superclass.instance_variable_get(:@__imported_classboxes__)
+    end
+    super(binding(1))
+  end
+end
+
+class Classbox
+  def __opened__
+    b = binding(1)
+    import_to(b)
+    super(b)
+  end
+end
diff --git a/object.c b/object.c
index e2ed9e8..ba1a20b 100644
--- a/object.c
+++ b/object.c
@@ -2535,6 +2535,7 @@ Init_Object(void)
     rb_define_private_method(rb_cModule, "method_added", rb_obj_dummy, 1);
     rb_define_private_method(rb_cModule, "method_removed", rb_obj_dummy, 1);
     rb_define_private_method(rb_cModule, "method_undefined", rb_obj_dummy, 1);
+    rb_define_private_method(rb_cModule, "__opened__", rb_obj_dummy, 0);
 
     rb_define_method(rb_mKernel, "nil?", rb_false, 0);
     rb_define_method(rb_mKernel, "===", rb_equal, 1);
-- 
1.7.0.4

